import os
import ast
import sys
import json
import inspect
import pathlib
import warnings
import functools
import importlib
from pprint import pformat
from . import stores
from . import exceptions as exc
from .utils import branches, getenv, parse_paths
from .utils import log, logd, sxpr_to_python, python_to_sxpr

# TODO multi-get to reduce reads via context manager maybe?

try:
    import yaml  # FIXME DANGERZONE :/
except ImportError as e:
    pass


def crpath(__file__, name):
    """ for use in combination with configure e.g.
        `auth = oa.configure(oa.cpath(__file__))`
        assuming that `auth-config.py` is the name
        of your auth config file """
    # NOTE: never set a default value for name
    # because the call to this function should
    # document the name of the file in a way that
    # is easy to see locally
    return pathlib.Path(__file__).parent / name


def configure(auth_config_path, include=tuple()):
    """ hrm """
    return AuthConfig(auth_config_path, include=include)


def configure_here(name, calling__name__, include=tuple()):
    calling_module = importlib.import_module(calling__name__)
    calling__file__ = calling_module.__file__
    return AuthConfig._from_relative_path(calling__file__,
                                          name,
                                          include=include,
                                          calling_module=calling_module)


def configure_relative(name, include=tuple()):
    """ WARNING very bad performance due to using inspect """
    warnings.warn(f'configure_relative is deprecated please switch to configure_here',
                    DeprecationWarning,
                    stacklevel=2)
    stack = inspect.stack(0)
    s1 = stack[1]
    calling__file__ = s1.filename
    calling_module = inspect.getmodule(s1.frame)
    return AuthConfig._from_relative_path(calling__file__,
                                          name,
                                          include=include,
                                          calling_module=calling_module)


class DecoBase:
    def environ(self, set_envar, from_name, when=True):
        """ in order to use orthauth with other projects that are not aware of
            its existence and that make use of environment variables orthauth
            can set the envar that the other project expects using a value from
            an orthauth managed auth store """
        raise NotImplementedError('TODO')

    def tangential_init(self, inject_value, with_name, when=True, after=False, method='get'):
        """ tangential decorator that defaults to atInit since it is a common use case

            note that this can be used to monkey patch classes as well """
        atInit = 'after' if after else True
        return self.tangential(inject_value, with_name, when=when, atInit=atInit, method=method)

    def tangential(self, inject_value, with_name, when=True, asProperty=False, atInit=False, method='get'):
        """ class decorator
            the tangential auth decorator makes a name available to
            all instances of a class from class creation time, this
            this is not fully orthogonal, but makes it easier to
            separate the logic of an API from the logic of its auth

            In some cases this can be used to overwrite an auth variable
            in a class from a project that is unaware that orthauth exists.

            atInit='after' -> bind the value after the original __init__
            instead of before, useful in cases where the value is set
            during __init__ """

        if not when:
            # FIXME not ready
            return lambda cls: cls

        if asProperty and atInit:
            raise ValueError('asProperty and atInit are mutually exclusive')

        @property
        def tangential_property(self, outer_self=self, name=with_name):
            return getattr(outer_self, method)(name)

        def cdecorator(cls=None, asProperty=asProperty, atInit=atInit, method=method):
            if asProperty and atInit:
                raise ValueError('asProperty and atInit are mutually exclusive')

            if cls is None:
                if asProperty:
                    def inner_cdec(icls):
                        setattr(icls, inject_value, tangential_property)
                        return icls

                    return inner_cdec
                elif atInit:
                    if atInit == 'after':
                        def inner_cdec(icls):
                            cls__init__ = cls.__init__
                            @functools.wraps(cls__init__)
                            def __init__(inner_self, *args, **kwargs):
                                cls__init__(inner_self, *args, **kwargs)
                                setattr(inner_self, inject_value, getattr(self, method)(with_name))

                            setattr(icls, '__init__', __init__)
                            return icls
                    else:  # life-without-macros
                        def inner_cdec(icls):
                            cls__init__ = cls.__init__
                            @functools.wraps(cls__init__)
                            def __init__(inner_self, *args, **kwargs):
                                setattr(inner_self, inject_value, getattr(self, method)(with_name))
                                cls__init__(inner_self, *args, **kwargs)

                            setattr(icls, '__init__', __init__)
                            return icls

                    return inner_cdec
                else:
                    def inner_cdec(icls):
                        setattr(icls, inject_value, getattr(self, method)(with_name))
                        return icls

                    return inner_cdec

            elif asProperty:
                setattr(cls, inject_value, tangential_property)

            elif atInit:
                cls__init__ = cls.__init__
                cls._cls__init__ = cls__init__  # TODO do need this even if we are forced to subclass :/ sigh, yeah, if you need multiple per ... the decorator has to return a different class, and it cant decorate
                if atInit == 'after':
                    @functools.wraps(cls__init__)
                    def __init__(inner_self, *args, **kwargs):
                        cls__init__(inner_self, *args, **kwargs)
                        setattr(inner_self, inject_value, getattr(self, method)(with_name))
                else:
                    @functools.wraps(cls__init__)
                    def __init__(inner_self, *args, **kwargs):
                        setattr(inner_self, inject_value, getattr(self, method)(with_name))
                        cls__init__(inner_self, *args, **kwargs)

                setattr(cls, '__init__', __init__)

            else:
                setattr(cls, inject_value, getattr(self, method)(with_name))

            return cls

        return cdecorator


class ConfigBase:

    def __new__(cls, path, include=tuple()):
        if isinstance(path, str):
            path = pathlib.Path(path)

        self = super().__new__(cls)
        try:
            key = path if path is None else path.suffix
            self._load_type = {None: self._load_runtime,
                               '.json': self._load_json,
                               '.py': self._load_python,
                               '.sxpr': self._load_sxpr,
                               '.yaml': self._load_yaml}[key]
        except KeyError as e:
            raise exc.UnsupportedConfigLangError(path.suffix) from e

        self._path = path

        if include:  # unqualified include ... in theory we could nest everything
            if isinstance(include, cls):
                include = include,

            inc = []
            avs = [self.get_blob('auth-variables')]
            for ipath in include:
                if isinstance(ipath, cls):
                    ic = ipath
                else:
                    ic = cls(ipath)

                inc.append(ic)
                iav = ic.get_blob('auth-variables')
                avs.append(iav)

            bads = set()
            for av in avs:
                others = set(k for a in avs if a != av for k in a)
                bads |= set(k for k in av if k in others) | set(k for k in others if k in av)

            if bads:
                # FIXME because the path variables are not expanded before being compared
                # this can cause false positives, where '../here' appearing twice when the
                # included configs start in different locations
                raise exc.VariableCollisionError(f'{bads}')

            self._include = tuple(inc)

        else:
            self._include = tuple()

        # XXX there is a tradeoff here, if we defer loading, then we
        # can start up faster, however, if there is an error in the
        # config, then things become very confusing and hard to debug
        # therefor on the balance we are better off detecting errors
        # in configs at load time so that it is clear what went wrong
        # we might be able to improve this a tiny bit by keeping a
        # single file that records paths and modified times for
        # configs so that we only have to read a single file and stat
        # the rest which should be faster than what we do right now,
        # however that is a micro optimization at this point, however
        # it may be worth it as the number of configs in use increases
        if self._path is not None:
            self.load()

        return self

    @classmethod
    def runtimeConfig(cls, blob):
        # FIXME pretty sure __init__ is not running here
        self = ConfigBase.__new__(cls, None)
        #self._path = None  # FIXME vs
        # /dev/null/this/path/does/not/exist/but/you/could/escape/with/enough/..
        self._runtime_config = blob
        return self  # NOTE init is complete if runtimeConfig is not implemented on the subclass

    def load_type(self):
        cn = self.__class__.__name__
        warnings.warn(f'{cn}.load_type is deprecated please switch to {cn}.load',
                      DeprecationWarning,
                      stacklevel=2)
        return self.load()

    def _mtime_changed(self):
        """ Check if the modified time of self._path has changed.
            Still hits the disk but can be used to avoid force a
            full read of the file when there are no changes. """

        if self._path is not None:  # no path, no disk, no worries
            mtime = self._path.stat().st_mtime
            if not hasattr(self, '_last_mtime'):
                last_mtime = self._last_mtime = mtime
            else:
                last_mtime, self._last_mtime = self._last_mtime, mtime

            return mtime != last_mtime

    def load(self):
        if not hasattr(self, '_blob') or self._mtime_changed():
            msg = (f'The config at {self._path!r} '
                   'should not be empty!')
            empty_error = exc.EmptyConfigError(msg)

            try:
                _blob = self._load_type()
            except json.decoder.JSONDecodeError as e:
                raise empty_error from e
            except SyntaxError as e:
                if (e.msg in ('unexpected EOF while parsing', 'invalid syntax') and
                    e.offset == 0):
                    raise empty_error from e
                else:
                    raise e

            if _blob is None:
                raise empty_error

            self._blob = _blob

        return self._blob

    def dump(self, config):
        """ Don't use this ... seriously """
        format = self._path.suffix.strip('.')
        string = self._dump(config, format)
        with open(self._path, 'wt') as f:
            f.write(string)

    def _dump(self, config, format):
        if format == 'json':
            return json.dumps(config, indent=2, sort_keys=True)
        elif format == 'py':
            return pformat(config)
        elif format == 'sxpr':
            # XXX expect failure for any complex types
            return python_to_sxpr(config)
        elif format == 'yaml':
            try:
                return yaml.dump(config, default_flow_style=False)
            except NameError as e:
                msg = (f'Module yaml was not found while tyring to load {self._path}\n'
                       'To resolve this issue reinstall orthauth with yaml '
                       'support enabled\n or reformat your config to .py')
                raise ModuleNotFoundError(msg) from e
        else:
            raise NotImplementedError(f'serialization to {format!r} is not ready')

    def get_blob(self, *names, fail=False):
        if not names:
            raise TypeError('names is a required argument')

        current = self.load()

        for name in names:
            current = current[name]

        if fail and isinstance(current, dict):
            raise ValueError(f'Your config path is incomplete.')

        return current

    def _load_string(self, string, format):
        loadf = {'json': json.loads,
                 'py': ast.literal_eval,
                 'sxpr': sxpr_to_python,
                 'yaml': yaml.safe_load,}[format]
        return loadf(string)

    def _load_runtime(self):
        return self._runtime_config

    def _load_json(self):
        with open(self._path, 'rt') as f:
            return json.load(f)

    def _load_python(self):
        """ python configuraiton files should be only a single dict literal """
        if (hasattr(self, '_calling_module') and
            self._path.suffix == '.py' and
            not self._path.exists()):
            # there are certain ways of installing a python package that
            # stick an auth-config.py file in a binary blob so we need to
            # do this little dance to guess the right module name and
            # import it including files with dashes in their name
            p = pathlib.Path(self._calling_module.__file__).parent
            prt = self._path.relative_to(p)
            modpath = '.'.join([p.name, *prt.with_suffix('').as_posix().split('/')])
            module = importlib.import_module(modpath)
            source = inspect.getsource(module)
            return ast.literal_eval(source)

        with open(self._path, 'rt') as f:
            return ast.literal_eval(f.read())

    def _load_sxpr(self):
        with open(self._path, 'rt') as f:
            return sxpr_to_python(f.read())

    def _load_yaml(self):
        with open(self._path, 'rt') as f:
            try:
                return yaml.safe_load(f)
            except NameError as e:
                msg = (f'Module yaml was not found while tyring to load {self._path}\n'
                       'To resolve this issue reinstall orthauth with yaml '
                       'support enabled\n or reformat your config to .py')
                raise ModuleNotFoundError(msg) from e

    @staticmethod
    def _envars(auth_variable_value):
        envars = []
        # env
        for evkey in ('environment-variables', 'env-vars', 'envars'):
            # FIXME do we allow multiple keys like this or error if
            # more than one is detected?
            if evkey in auth_variable_value:
                if evkey in auth_variable_value:
                    ev = auth_variable_value[evkey]
                    if isinstance(ev, str):
                        envars += ev.split(' ')
                    elif isinstance(ev, list):
                        envars += ev
                    else:
                        # this is actually a type error because it means
                        # something is wrong with the implementation not
                        # with what the user did
                        raise TypeError(f'unsupported type {type(ev)}\n{ev}')

        return envars

    @staticmethod
    def _auth_paths(auth_variable_value):
        if 'config' in auth_variable_value:
            if [_ for _ in ('path', 'paths', 'paths-nested') if _ in auth_variable_value]:
                msg = ('can only have config or path, not both')
                raise exc.BadAuthConfigFormatError(msg)

            return []

        raw_paths = []
        # paths
        if 'path' in auth_variable_value:
            raw_paths += [auth_variable_value['path']]

        if 'paths' in auth_variable_value:
            raw_paths += auth_variable_value['paths']

        if 'paths-nested' in auth_variable_value:
            raw_paths += list(branches(auth_variable_value['paths-nested']))

        try:
            paths = list(parse_paths(raw_paths))
        except exc.VariableNotDefinedError as e:
            raise e  # TODO message about where missing from

        return paths

    @staticmethod
    def _single_alt_configs(auth_variable_value):
        if 'config' in auth_variable_value:
            if 'path' in auth_variable_value:
                raise TypeError('can only have config or path, not both')

            return auth_variable_value

    def _pathit(self, path_string):
        return self._pathit_relative(self._path, path_string)

    def _pathit_relative(self, relative_base_child, path_string):
        if not isinstance(path_string, pathlib.Path):
            if '{:' in path_string:  # special paths  # FIXME vs startswith
                path_string = path_string.replace('{:', '{')
                path_string = path_string.format(**self._config_vars)

            path = pathlib.Path(path_string)
        else:
            path = path_string

        if path.parts[0] == ('~'):
            path = path.expanduser()

        if not path.is_absolute():
            if relative_base_child is None:
                raise exc.NoBasePathError(f'Cannot resolve relative path {path} '
                                          'from runtime config that lacks a path.')
            else:
                return relative_base_child.parent / path

        if '..' in path.parts:
            path = path.resolve()

        return path

    @property
    def _config_vars(self):
        return self._make_config_vars()

    @staticmethod
    def _make_config_vars():
        """ currently supported config vars

            cwd
            prefix
            user-cache-path
            user-config-path
            user-data-path
            user-log-path
        """
        if os.name != 'nt':
            if sys.platform == 'darwin':
                ucp = '~/Library/Application Support'
                udp = '~/Library/Application Support'
                uchp = '~/Library/Caches'
                ulp = '~/Library/Logs'
            else:
                ucp = '~/.config'
                udp = '~/.local/share'
                uchp = '~/.cache'
                ulp = '~/.cache/log'
        else:
            ucp = '~/AppData/Local'
            udp = ucp
            uchp = ucp
            ulp = '~/AppData/Local/Logs'

        return {'cwd': pathlib.Path.cwd(),
                'prefix': sys.prefix,
                'user-cache-path': uchp,
                'user-config-path': ucp,
                'user-data-path': udp,
                'user-log-path': ulp,
        }

    @staticmethod
    def _config_vars_expanded():
        return {k:pathlib.Path(v).expanduser()
                for k, v in AuthConfig._make_config_vars().items()}


class AuthConfig(DecoBase, ConfigBase):  # FIXME this is more a schema?
    """ Object representation of a static configuration file
    that lives in a repository and that changes only when some
    change needs to be made to a decorator in the code base that
    needs authenticated access.

    This is the primary api entry point for orthauth.
    """

    @classmethod
    def _from_relative_path(cls, calling__file__, name, include=tuple(), calling_module=None):
        self = super().__new__(cls, pathlib.Path(calling__file__).parent / name, include=include)
        self._calling_module = calling_module
        self._user_config = UserConfig(self)
        return self

    def __new__(cls, path, include=tuple()):
        self = super().__new__(cls, path, include=include)
        self._user_config = UserConfig(self)
        return self

    @classmethod
    def runtimeConfig(cls, *blobs):
        blob, *child_blobs = blobs
        self = super().runtimeConfig(blob)
        self._user_config = UserConfig.runtimeConfig(self, *child_blobs)
        return self

    @property
    def user_config(self):
        uc = self._user_config
        ac = uc.alt_config
        if ac:
            uc = ac
        
        return uc

    @property
    def dynamic_config(self):
        cn = self.__class__.__name__
        warnings.warn(f'{cn}.dynamic_config is deprecated please switch to {cn}.user_config',
                      DeprecationWarning,
                      stacklevel=2)
        return self.user_config

    @property
    def user_config_paths(self):
        return [self._pathit(path_string) for path_string in
                self.get_blob('config-search-paths')]

    @property
    def user_config_path(self):
        ucps = self.user_config_paths
        for path in ucps:
            if path.exists():
                return path

        # if no config exists automatically create the default
        ucp = ucps[0]  # you MUST have a user config path
        self.write_user_config(ucp=ucp)
        return ucp

    def load(self):
        if not hasattr(self, '_blob'):
            # auth configs are static so it is ok to cache them
            self._blob = super().load()

        return self._blob

    def get_blob(self, *names):
        blob = super().get_blob(*names)
        if blob is None:
            raise KeyError(f'{names}')

        return blob

    def __call__(self, cls_or_function):
        raise NotImplementedError

    def _check(self):
        """ make sure restrictions on config structure are satisfied """
        # error on unexpected keys to prevent forgetting variables:
        raise NotImplementedError

    def _make_user(self):
        return {'auth-stores': {'secrets': {'path': '{:user-config-path}/orthauth/secrets.yaml'}},
                'auth-variables': {var:None for var in self.get_blob('auth-variables')}}

    def _serialize_user_config(self, format):
        config = self._make_user()
        return self._dump(config, format)

    def write_user_config(self, *, format=None, ucp=None):
        # NOTE user config cannot write itself
        ucps = self.user_config_paths
        for _u in ucps:  # if any config already exists exit
            if _u.exists():
                raise exc.ConfigExistsError('{_u}')

        if ucp is None:
            ucp = self.user_config_path

        if format is not None:
            ucp = ucp.with_suffix('.' + format)
            if ucp not in ucps:
                # FIXME we do need to support .* or {yaml,py,lisp,json}
                raise TypeError(f'{ucp} not one of the expected formats {ucps}')
        else:
            format = ucp.suffix.strip('.')

        if not ucp.parent.exists():
            ucp.parent.mkdir(parents=True)

        string = self._serialize_user_config(format)
        with open(ucp, 'wt') as f:
            f.write(string)

    def get_list(self, variable_name):
        """ if you know a variable holds a list use this """
        var = self.get(variable_name, for_list=True)
        if var is None:
            return []
        else:
            return var

    def get_path(self, variable_name):
        """ if you know a variable holds a path use this to autoconvert """
        var = self.get(variable_name, for_path=True)
        # NOTE: right now if default is explicit for path comes back as a string
        # whereas a implicit default comes back as a Path already
        if var == [None]:
            return

        if isinstance(var, list):
            if not var:
                return

            first = None
            warnings = []
            for v in var:
                if v is None:
                    if warnings:
                        for w in warnings:
                            log.warning(w)
                    msg = f'{variable_name} in {self._path}'
                    raise exc.SomethingWrongWithVariableInConfig(msg)

                p = self._pathit(v)
                if first is None:
                    first = p

                if p.exists():
                    return p
                else:
                    msg = (f'path for {variable_name} does not exist, '
                           'did you enter a path: as a value?')
                    # never log v or any values derived from v as they
                    # might be secrets, the user will have to figure out
                    # where they messed up by looking at the configs directly
                    warnings.append(msg)
            else:
                return first

        else:
            return self._pathit(var)

    def get_default(self, variable_name, *args, **kwargs):
        av = self.get_blob('auth-variables')
        auth_variable_value = av[variable_name]

        defaults = []
        if not isinstance(auth_variable_value, dict):
            if isinstance(auth_variable_value, list):
                if ('for_path' not in kwargs or not kwargs['for_path']):
                    log.warning(f'attempting to get a default value for {variable_name} '
                                'that is a list did you want get_path?')

                defaults.extend(auth_variable_value)
            else:
                defaults.append(auth_variable_value)

            auth_variable_value = {}
        elif 'default' in auth_variable_value:
            d = auth_variable_value['default']
            if isinstance(d, list):
                defaults.extend(d)
            else:
                defaults.append(d)

        if defaults:
            return defaults[0]

    def get_path_default(self, variable_name):
        d = self.get_default(variable_name, for_path=True)
        if d is not None:
            return self._pathit(d)

    def get(self, variable_name, *args, **kwargs):
        """ look up the value of a variable name from auth store or config """
        for_list = 'for_list' in kwargs and kwargs['for_list']
        for_path = 'for_path' in kwargs and kwargs['for_path']
        av = self.get_blob('auth-variables')
        if variable_name not in av:
            error = None
            for i in self._include:
                try:
                    return i.get(variable_name, **kwargs)
                except KeyError as e:  # is indeed a key error
                    error = e
            else:
                if error is not None:
                    raise error

        try:
            user_variable_value = self.user_config.get_blob('auth-variables', variable_name)
            # we use get_blob to reduce the number of times we read from disk, we dereference
            # the value in secrets below if there is one, but we defer that until later, so
            # do not call self.user_config.get() at this point
            if user_variable_value is None:
                user_variable_value = {}
                f1 = TypeError(f'Value of {variable_name} is None'
                               f' in user config {self.user_config._path}')
            else:
                f1 = False
        except KeyError as e:
            user_variable_value = {}
            f1 = e

        try:
            auth_variable_value = av[variable_name]
            f2 = False
        except KeyError as e:
            auth_variable_value = {}
            f2 = e

        if f1 and f2:
            raise f2 from f1

        defaults = []
        if not isinstance(user_variable_value, dict):
            if isinstance(user_variable_value, list):
                if for_list:
                    user_variable_value = [user_variable_value]
                elif not for_path:
                    log.warning(f'attempting to get a default value for {variable_name} '
                                'that is a list did you want get_path?')
                else:
                    user_variable_value = [
                        self.user_config._pathit(p)
                        for p in user_variable_value if p is not None]

                defaults.extend(user_variable_value)
            else:
                if for_path and user_variable_value is not None:
                    user_variable_value = self.user_config._pathit(user_variable_value)

                defaults.append(user_variable_value)

            user_variable_value = {}

        # there is a limited use case for allowing users to set a value, if they also
        # want to define custom user set environment variables
        elif 'default' in user_variable_value:
            d = user_variable_value['default']
            if isinstance(d, list):
                if for_list:
                    d = [d]
                elif not for_path:
                    log.warning(f'attempting to get a default value for {variable_name} '
                                'that is a list did you want get_path?')
                else:
                    d = [self.user_config._pathit(p) for p in d if p is not None]

                defaults.extend(d)

            else:
                if for_path and d is not None:
                    d = self.user_config._pathit(d)

                defaults.append(d)

        if not isinstance(auth_variable_value, dict):
            if isinstance(auth_variable_value, list):
                if for_list:
                    auth_variable_value = [auth_variable_value]
                elif not for_path:
                    log.warning(f'attempting to get a default value for {variable_name} '
                                'that is a list did you want get_path?')
                else:
                    auth_variable_value = [self._pathit(p)
                                  for p in auth_variable_value if p is not None]

                defaults.extend(auth_variable_value)
            else:
                if for_path and auth_variable_value is not None:
                    auth_variable_value = self._pathit(auth_variable_value)

                defaults.append(auth_variable_value)

            auth_variable_value = {}

        elif 'default' in auth_variable_value:
            d = auth_variable_value['default']
            if isinstance(d, list):
                if for_list:
                    d = [d]
                elif not for_path:
                    log.warning(f'attempting to get a default value for {variable_name} '
                                'that is a list did you want get_path?')
                else:
                    d = [self._pathit(p) for p in d if p is not None]

                defaults.extend(d)

            else:
                if for_path and d is not None:
                    d = self._pathit(d)

                defaults.append(d)

        envars = self._envars(user_variable_value)
        envars += self._envars(auth_variable_value)
        paths = self._auth_paths(user_variable_value)  # this is not expected to raise a KeyError
        paths += self._auth_paths(auth_variable_value)

        bads = self._single_alt_configs(auth_variable_value)  # for error purposes only
        if bads:
            msg = ('auth-configs should never define single atl configs\n'
                   f'{bads} in {self._path}')
            raise exc.BadAuthConfigFormatError(msg)

        alt = self._single_alt_configs(user_variable_value), variable_name, for_path

        if for_path:
            get_uc = self.user_config._get_path
            def get_default(d):
                return d
        elif for_list:
            get_uc = self.user_config._get
            def get_default(d):
                return d[0]

        else:
            get_uc = self.user_config._get
            def get_default(d):
                if len(d) == 1 and d[0] is None:
                    return
                else:
                    return d[0]

        for f, v in zip((getenv,
                         get_uc,
                         self.user_config._gsac_wrap,
                         get_default),
                        (envars, paths, alt, defaults)):
            if v:
                SECRET = f(v)
                if SECRET is not None:
                    return SECRET


class UserConfig(ConfigBase):
    """ Same structure as secrets, but all information
    is identifying, not authenticating

    There is a strong possibility that this will only be a single
    level without any trees to simplify the interaction between
    the Auth and User configs

    WARNING: if you use this to set authentication endpoints
    then make sure a malicious party cannot change the endpoint
    to steal credentials that you will be sending to that endpoint
    """

    @classmethod
    def _from_user_alt_config(cls, path, auth_config, rename=None):
        # TODO rename
        self = super().__new__(cls, path)
        if rename:
            self.__rename = rename
            self.load = self._rename_load

        self.auth_config = auth_config
        return self

    def _rename_load(self):
        blob = super().load()
        av = blob['auth-variables']
        log.debug(self.__rename)
        [log.debug(k) for k in av.keys()]
        for to, frm in self.__rename.items():
            if frm in av:
                av[to] = av.pop(frm)

        [log.debug(k) for k in av.keys()]
        return blob

    def __new__(cls, auth_config):
        path = auth_config.user_config_path
        self = super().__new__(cls, path)
        self.auth_config = auth_config
        return self

    @classmethod
    def runtimeConfig(cls, auth_config, *blobs):
        if not blobs:
            return UserConfig.__new__(cls, auth_config)
        else:
            blob, *child_blobs = blobs
            self = super().runtimeConfig(blob)
            self.auth_config = auth_config
            # TODO auth store level?
            if child_blobs:
                self._secrets_runtime = stores.Runtime(child_blobs[0])

            return self

    def _auth_store(self, type_):
        try:
            return {
                'secrets': self._secrets,
                'authinfo': self._authinfo,
                'mypass': self._mypass,
                'ssh-config': self._ssh_config,
            }[type_]
        except KeyError as e:
            raise exc.UnknownAuthStoreType(type_)

    def _authinfo(self, blob):
        return stores.Authinfo(self._blob_path(blob))

    def _get_store(self, store, blob):
        cache_name = f'_c_{store.__name__}'
        if hasattr(self, cache_name):
            return getattr(self, cache_name)

        path = self._blob_path(blob)
        try:
            store_obj = store(path)
            setattr(self, cache_name, store_obj)
            return store_obj
        except FileNotFoundError as e:
            msg = (f'In the config located at: {self._path!r}\n'
                   f'the following path does not exist: {path!r}\n'
                   f'the raw value from the config was: {blob["path"]}')
            err = exc.SomethingWrongWithVariableInConfig(msg)
            raise err from e

    def _secrets(self, blob):
        try:
            return self._get_store(stores.Secrets, blob)
        except exc.NoBasePathError as e:
            if hasattr(self, '_secrets_runtime'):
                return self._secrets_runtime
            else:
                raise e

    def _mypass(self, blob):
        return self._get_store(stores.Mypass, blob)

    def _ssh_config(self, blob):
        return self._get_store(stores.SshConfig, blob)

    def _blob_path(self, blob):
        return self._pathit(blob['path'])

    def get(self, variable_name, *args, **kwargs):
        """ look up the value of a variable name from auth store or config """
        # XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX this is NOT CALLED by AuthConfig.get !!!!!!!!!!!!!!!
        # ah the problem of interleveing values from sources of different rank ...
        for_list = 'for_list' in kwargs and kwargs['for_list']
        for_path = 'for_path' in kwargs and kwargs['for_path']
        auth_variable_value = self.get_blob('auth-variables', variable_name)
        if auth_variable_value is None:
            raise KeyError(variable_name)

        defaults = []
        if not isinstance(auth_variable_value, dict):
            if isinstance(auth_variable_value, list):
                if for_list:
                    auth_variable_value = [auth_variable_value]
                elif not for_path:
                    log.warning(f'attempting to get a default value for {variable_name} '
                                'that is a list did you want get_path?')

                defaults.extend(auth_variable_value)
            else:
                defaults.append(auth_variable_value)

            auth_variable_value = {}

        envars = self._envars(auth_variable_value)
        paths = self._auth_paths(auth_variable_value)
        alt = self._single_alt_configs(auth_variable_value), variable_name, for_path

        if for_path:
            def get_default(d):
                return d
        else:
            def get_default(d):
                if len(d) == 1 and d[0] is None:
                    return
                else:
                    return d[0]

        for f, v in zip((getenv,
                         self._get_path if for_path else self._get,
                         self._gsac_wrap,
                         get_default),
                        (envars, paths, alt, defaults)):
            if v:
                SECRET = f(v)
                if SECRET is not None:
                    return SECRET

    def _get_path(self, paths):
        store_path, value = self._get_with_path(paths)
        return self._pathit_relative(store_path, value)

    def _get(self, paths):
        _, value = self._get_with_path(paths)
        return value

    def _get_with_path(self, paths):
        secrets = self.secrets
        errors = []
        for names in paths:
            auth_store = self.path_source(*names)  # FIXME perf issue incoming
            try:
                return auth_store._path, auth_store(*names)
            except exc.SecretError as e:
                errors.append(e)
                logd.error(f'broken path {names}')
                if auth_store != secrets:
                    try:
                        return secrets._path, secrets(*names)
                    except exc.SecretError as e:
                        errors.append(e)

        if errors:
            raise exc.SecretError(f'{[e.args[0] for e in errors]}') from errors[-1]

    def _gsac_wrap(self, args):
        return self._get_single_alt_config(*args)

    def _get_single_alt_config(self, blob, variable, for_path=False):
        # FIXME relative paths get funky here
        if blob:
            config = blob['config']
            path = self._pathit(config)
            rename = blob.get('rename', None)
            if rename is not None:
                variable = rename

            auc = self._from_user_alt_config(path, self.auth_config)
            return auc.get(variable, for_path=for_path)

    @property
    def _alt_config_path(self):
        blob = self.get_blob('alt-config')
        if not isinstance(blob, str):
            raise TypeError('only a single alt config is allowed')
        else:
            return self._pathit(blob)

    @property
    def _rename(self):
        blob = self.get_blob('rename')
        if not isinstance(blob, str):
            raise TypeError('only a single alt config is allowed')
        else:
            return self._pathit(blob)

    @property
    def alt_config(self):
        if hasattr(self, '_c_alt_config'):
            return self._c_alt_config

        try:
            path = self._alt_config_path
            _test = self.load()
            _test.pop('alt-config')
            rename = _test.pop('rename', None)
            if _test:
                msg = ('configs with top level alt-config may '
                       f'have only a rename section\n{_test}')
                raise exc.BadAuthConfigFormatError(msg)

            self._c_alt_config = self._from_user_alt_config(path,
                                                            self.auth_config,
                                                            rename)
            return self._c_alt_config
        except KeyError:
            pass

    @property
    def secrets(self):
        """ default failover for paths without explicit management """
        try:
            return self._secrets(self.get_blob('auth-stores', 'secrets'))
        except:
            if hasattr(self, '_secrets_runtime') and self.get_blob('auth-stores', 'runtime'):
                return self._secrets_runtime
            else:
                raise

    @property
    def _path_sources(self):
        # FIXME from type reads twice, should only read once
        try:
            path_data = self.get_blob('path-sources')
        except KeyError:
            return {}

        store_data = self.get_blob('auth-stores')
        out = {}
        for raw_path, type_ in path_data.items():
            try:
                names = tuple(next(parse_paths([raw_path])))
            except exc.VariableNotDefinedError as e:
                logd.error(f'variable missing in {self._path}')
                raise e  # TODO message about where missing from

            blob = store_data[type_]  # key error here is ok and expected but need good messaging
            auth_store = self._auth_store(type_)(blob)
            out[names] = auth_store

        return out

    def path_source(self, *names):
        sources = self._path_sources
        search = names
        while search:
            if search in sources:
                auth_store = sources[tuple(names)]
                return auth_store
            search = search[:-1]

        else:
            return self.secrets





class OrthauthError(Exception):
    """ base class for orthauth errors """


class SecretError(OrthauthError):
    """ base clasee for errors related to secrets files """


class SecretAsKeyError(SecretError):
    """ Using a secret as a key in the path to another secret
        is a recipe for secrets getting committed to code. """


class SecretPathError(SecretError):
    """ Something is wrong with a config/auth/secret path spec
        it might be too long, it might be too short, it might
        contain a secret, etc. """


class SecretEmptyError(SecretError):
    """ Having empty paths in secret files is bad practice.
        If we detect the issue at runtime we raise this error. """


class UnknownAuthStoreType(OrthauthError):
    """ This type of authentication store is unknown """


class UnsupportedConfigLangError(OrthauthError):
    """ Using a secret as a key in the path to another secret
        is a recipe for secrets getting committed to code. """


class VariableNotDefinedError(OrthauthError):
    """ Variable is not know to any config file """


class ConfigExistsError(OrthauthError):
    """ Config file already exists, so don't overwrite it """


class EmptyConfigError(OrthauthError):
    """ Config file is empty! """


class VariableCollisionError(OrthauthError):
    """ a variable is colliding between configs """


class BadAuthConfigFormatError(OrthauthError):
    """ that's some baaaaad data """


class SomethingWrongWithVariableInConfig(OrthauthError):
    """ a variable in a config is bad """


class NoBasePathError(OrthauthError):
    """ tried to resolve a path relative to None usually because
        a runtime config value was not an absolute path"""


class BadFilePermissionsError(OrthauthError):
    """ file or folder has bad/dangerous/insecure settings """





import os
import logging
from . import exceptions as exc

try:
    from sxpyr import sxpyr
except ImportError as e:
    pass


def python_to_sxpr(thing):
    pl = sxpyr.python_to_sxpr(thing)
    return pl._print(sxpyr.print_plist)


def sxpr_to_python(string):
    parse_plist = sxpyr.configure(**sxpyr.conf_plist)  # FIXME error on lack of **
    read_plist = sxpyr.conf_read(parse_plist, sxpyr.WalkPl)
    def cf(ast):
        sxpyr  # yay scoping rules
        if isinstance(ast, sxpyr.PList):
            v = [_.caste(cf) if hasattr(_, 'caste') else _
                 for _ in ast.value]  # recurse
            if v:
                return sxpyr.plist_to_dict(v)
        elif isinstance(ast, sxpyr.List):
            # XXX this isn't the ast at this point, it is the DataType layer
            # sometimes an Ast leaks through ... sigh inhomogenaity in the IR
            return [_.caste(cf) if hasattr(_, 'caste') else cf(_)
                    for _ in ast.value]
        elif isinstance(ast, sxpyr.Ast):
            return ast.value
        else:
            return ast

    def _cf(ast):
        if False and isinstance(ast, sxpyr.PList):
            return sxpyr.plist_to_dict(ast.value)
        elif isinstance(ast, sxpyr.ListAbstract):
            return [_.caste(cf) for _ in ast.collect]
        elif isinstance(ast, sxpyr.Ast):
            return ast.value
        elif isinstance(ast, str):
            return ast
        else:
            raise NotImplementedError(f'asdf {ast!r}')
            return ast

    _raw = list(read_plist(string))
    if not _raw:
        # note that nil () -> python [] so None is python convention
        # however because this is in orthauth we just raise the empty
        # config error and cut out the middleman
        raise exc.EmptyConfigError()

    raw = _raw[0]
    asdf = raw.caste(cf)
    return asdf


def makeSimpleLogger(name, level=logging.INFO):
    # TODO use extra ...
    logger = logging.getLogger(name)
    logger.setLevel(level)
    ch = logging.StreamHandler()  # FileHander goes to disk
    fmt = ('[%(asctime)s] - %(levelname)8s - '
           '%(name)14s - '
           '%(filename)16s:%(lineno)-4d - '
           '%(message)s')
    formatter = logging.Formatter(fmt)
    ch.setFormatter(formatter)
    logger.addHandler(ch)
    return logger


log = makeSimpleLogger('orthauth')
logd = log.getChild('data')


def getenv(ordered_vars):
    for var in ordered_vars:
        if var in os.environ:
            value = os.environ[var]
            if value is not None:
                return value


def branches(tree):
    for name, subtree in tree.items():
        if subtree is None:
            yield [name]
        else:
            for subbranch in branches(subtree):
                yield [name, *subbranch]


def parse_paths(raw_paths):
    for elements in raw_paths:
        if isinstance(elements, str):
            elements = elements.split(' ')

        
        path = []
        for element in elements:
            if isinstance(element, int):
                element = str(element)

            path.append(element)

        yield path


class QuietTuple(tuple):
    """ read only doesn't print, repr, reduce etc. """
    def __add__(self, value):
        raise TypeError('NOPE')

    def __repr__(self):
        return '[secure]'

    def __str__(self):
        return '[secure]'

    def __reduce__(self):
        return (tuple, tuple())


class QuietDict(dict):
    """ read only doesn't print, repr, reduce etc. """
    def copy(self):
        return None

    def pop(self, key):
        return None

    def popitem(self, key):
        return None

    def update(self, value):
        return None

    def values(self):
        return QuietTuple(super().values())

    def __repr__(self):
        return '{secure}'

    def __str__(self):
        return '{secure}'

    def __reduce__(self):
        return (dict, tuple())



import unittest
import orthauth as oa
from orthauth import exceptions as exc
from .common import test_folder


class TestSecrets(unittest.TestCase):

    def setUp(self):
        auth = oa.configure(test_folder / 'auth-config-1.yaml')
        self.secrets = auth.user_config.secrets  # this is secrets-test-1.yaml

    def test_ANGRY(self):
        try:
            test = self.secrets('evil-path', 'DO NOT WANT')
            assert False, 'should have failed'
        except exc.SecretAsKeyError:
            pass

    def test_dissapointed(self):
        try:
            test = self.secrets('user', 'defined', 'path')
            assert False, 'should have failed'
        except exc.SecretPathError:
            pass

    def test_ANGRY_dissapointed(self):
        try:
            test = self.secrets('really-evil-path', 'paved', 'with')
            assert False, 'should have failed'
        except exc.SecretAsKeyError:
            pass

    def test_ANGRIER_dissapointed(self):
        try:
            test = self.secrets('really-evil-path', 'paved', 'utoh')
            assert False, 'should have failed'
        except exc.SecretAsKeyError:
            pass

    def test_null(self):
        try:
            test = self.secrets('you', 'see', 'nothing')
            assert False, 'should have failed'
        except exc.SecretEmptyError:
            pass



import pickle
import unittest
from orthauth.utils import branches, parse_paths, QuietDict, QuietTuple
from orthauth import exceptions as exc

tree = {
    1: {4: {5: {6: None}}},
    2: {7: None,
        8: None},
    3: None,
}


class TestBranches(unittest.TestCase):
    def test_branches(self):
        out = list(branches(tree))
        assert out == [
            [1, 4, 5, 6],
            [2, 7],
            [2, 8],
            [3],
        ]


paths = [
    'hello OK is a path',
    ['so', 'is', 'this'],
    [1, 2, 3],
    ['a', 1],
]


class TestParseAndExandPaths(unittest.TestCase):
    def test_parse(self):
        l = list(parse_paths(paths))
        assert l[0] == ['hello', 'OK', 'is', 'a', 'path']
        assert l[3] == ['a', '1']


class TestQuiet(unittest.TestCase):
    def test_quiet_dict(self):
        ref = {'oh': 'nose', 'her': 'nanomachines',}
        qd = QuietDict({'oh': 'nose',
                        'her': 'nanomachines',})

        f_r = (
            (str, lambda r: r == '{secure}'),
            (repr, lambda r: r == '{secure}'),
            ((lambda i: getattr(i, 'pop')('oh')), lambda r: r == None and qd == ref),
            ((lambda i: getattr(i, 'values')()), lambda r: r == QuietTuple(ref.values())),
            ((lambda i: getattr(i, 'popitem')('her')), lambda r: r == None and qd == ref),
            ((lambda i: getattr(i, 'update')({'weo': 'weo'})), lambda r: r == None and qd == ref),
            (pickle.dumps, lambda r: pickle.loads(r) == {}),
        )
        for f, r in f_r:
            assert r(f(qd))

    def test_quiet_tuple(self):
        qt = QuietTuple((1, 2, 3, 4))

        f_r = (
            (str, lambda r: r == '[secure]'),
            (repr, lambda r: r == '[secure]'),
            (pickle.dumps, lambda r: pickle.loads(r) == tuple()),
        )
        for f, r in f_r:
            assert r(f(qt))




import pathlib
import orthauth as oa

oa.utils.log.setLevel('DEBUG')

test_folder = pathlib.Path(__file__).parent / 'configs'

s1 = test_folder / 'secrets-test-1.yaml'
s1.chmod(0o0600)

s2 = (test_folder / '../secrets/secrets-2.yaml').resolve()
s2.chmod(0o0600)

se = (test_folder / 'secrets-empty.yaml').resolve()
se.chmod(0o0600)



import json
import yaml
from sxpyr import sxpyr
import pprint
import unittest
import orthauth as oa
from orthauth import exceptions as exc
from .common import test_folder


class TestFormats(unittest.TestCase):

    def _config(self, name):
        path = test_folder / name
        return oa.AuthConfig(path)

    def _do_test(self, auth):
        SECRET = auth.get('full-complexity-example')
        assert SECRET == 'oh-no-my-api-key-is-on-github-!', 'derp'

    def test_runtime(self):
        _source = self._config('auth-config-1.yaml')
        _ablob = _source.load()
        # FIXME user-config-path ??? more uniform with other behavior
        _ablob['config-search-paths'] = [test_folder / _ablob['config-search-paths'][0]]

        source = self._config('auth-config-1.yaml')
        ablob = source.load()
        ublob = source.user_config.load()

        with open(test_folder / 'secrets-test-1.yaml', 'rt') as f:  # please never do this irl
            sblob = yaml.safe_load(f)

        try:
            auth = oa.AuthConfig.runtimeConfig(ablob)
            assert False, 'should have failed due to non-relative-path'
        except exc.NoBasePathError:
            pass

        auth = oa.AuthConfig.runtimeConfig(_ablob)
        assert auth.get('default-example') == 42, 'deep thought required'
        auth = oa.AuthConfig.runtimeConfig(ablob, ublob)
        assert auth.get('test-after-init') == 'after-init', 'failure'
        auth = oa.AuthConfig.runtimeConfig(ablob, ublob, sblob)
        assert auth.get('oh-nose-her-api-keys') == 'DO NOT WANT', '( ͡° ͜ʖ ͡°)'

    def test_yaml(self):
        auth = self._config('auth-config-1.yaml')
        self._do_test(auth)

    def test_python(self):
        path = test_folder / 'auth-config-1.py'
        try:
            with open(test_folder / 'auth-config-1.yaml', 'rt') as f, open(path, 'wt') as o:
                d = yaml.safe_load(f)
                o.write(pprint.pformat(d))

            config = self._config(path.name)
            self._do_test(config)

        finally:
            if path.exists():
                path.unlink()

    def test_json(self):
        path = test_folder / 'auth-config-1.json'
        try:
            with open(test_folder / 'auth-config-1.yaml', 'rt') as f, open(path, 'wt') as o:
                d = yaml.safe_load(f)
                json.dump(d, o)

            config = self._config(path.name)
            self._do_test(config)

        finally:
            if path.exists():
                path.unlink()

    def test_sxpr(self):
        path = test_folder / 'auth-config-1.sxpr'
        try:
            with open(test_folder / 'auth-config-1.yaml', 'rt') as f, open(path, 'wt') as o:
                d = yaml.safe_load(f)
                o.write(oa.utils.python_to_sxpr(d))

            config = self._config(path.name)
            self._do_test(config)

        finally:
            if path.exists():
                path.unlink()


class TestEmptyAuthConfig(unittest.TestCase):

    _config = TestFormats._config

    def _do_test(self, suffix):
        path = (test_folder / 'auth-config-empty').with_suffix(suffix)
        try:
            with open(path, 'wt') as f:
                pass

            try:
                config = self._config(path.name)
                assert False, 'should have failed'
            except exc.EmptyConfigError:
                pass

        finally:
            if path.exists():
                path.unlink()

    def test_json(self):
        self._do_test('.json')

    def test_python(self):
        self._do_test('.py')

    def test_yaml(self):
        self._do_test('.yaml')

    def test_sxpr(self):
        self._do_test('.sxpr')


class TestEmptyUserConfig(TestEmptyAuthConfig):

    def _do_test(self, suffix):
        path = (test_folder / 'user-config-empty').with_suffix(suffix)

        _source = self._config('auth-config-1.yaml')
        _ablob = _source.load()
        _ablob['config-search-paths'] = [path]

        try:
            with open(path, 'wt') as f:
                pass

            try:
                auth = oa.AuthConfig.runtimeConfig(_ablob)
                #auth.get('default-example')  # needed if don't call self.load() in __new__
                assert False, 'should have failed due to empty user config'
            except exc.EmptyConfigError:
                pass

        finally:
            if path.exists():
                path.unlink()



import unittest
import orthauth as oa
from .common import test_folder


class TestDec(unittest.TestCase):
    def setUp(self):
        sc = test_folder / 'auth-config-1.yaml'
        self.auth = oa.AuthConfig(sc)

    def test_asProperty(self):
        @self.auth.tangential('api_key', 'test-as-property', asProperty=True)
        class Test:
            pass

        tv = 'apapi'
        test = Test()
        assert test.api_key == tv, [d for d in test.__dict__ if not d.startswith('__')]

    def test_asProperty_at_decoration_time(self):
        dec = self.auth.tangential('api_key', 'test-as-property')
        @dec(asProperty=True)
        class Test:
            pass

        tv = 'apapi'
        test = Test()
        assert test.api_key == tv, [d for d in test.__dict__ if not d.startswith('__')]

    def test_atInit(self):
        @self.auth.tangential('api_key', 'test-at-init', atInit=True)
        class Test:
            pass

        assert not hasattr(Test, 'api_key')
        tv = 'atapi'
        test = Test()
        assert test.api_key == tv, [d for d in test.__dict__ if not d.startswith('__')]

    def test_aInit_at_decoration_time(self):
        dec = self.auth.tangential('api_key', 'test-at-init')
        @dec(atInit=True)
        class Test:
            pass

        assert not hasattr(Test, 'api_key')
        tv = 'atapi'
        test = Test()
        assert test.api_key == tv, [d for d in test.__dict__ if not d.startswith('__')]

    def test_after_init(self):
        @self.auth.tangential('api_key', 'test-after-init', atInit='after')
        class Test:
            def __init__(self):
                self.api_key = 'YOU FELL FOR IT FOOL'

        assert not hasattr(Test, 'api_key')
        tv = 'after-init'
        test = Test()
        assert test.api_key == tv, [d for d in test.__dict__ if not d.startswith('__')]

    def test_after_init_adt(self):
        dec = self.auth.tangential('api_key', 'test-after-init')
        @dec(atInit='after')
        class Test:
            def __init__(self):
                self.api_key = 'YOU FELL FOR IT FOOL'

        assert not hasattr(Test, 'api_key')
        tv = 'after-init'
        test = Test()
        assert test.api_key == tv, [d for d in test.__dict__ if not d.startswith('__')]

    def test_after_init_adt_should_fail(self):
        dec = self.auth.tangential('api_key', 'test-after-init')
        @dec(atInit=True)
        class Test:
            def __init__(self):
                self.api_key = 'YOU FELL FOR IT FOOL'

        assert not hasattr(Test, 'api_key')
        tv = 'after-init'
        test = Test()
        assert test.api_key == tv, [d for d in test.__dict__ if not d.startswith('__')]

    def test_after_init_adt_unset_fail(self):
        dec = self.auth.tangential('api_key', 'test-after-init', atInit=True)
        @dec(atInit=False)
        class Test:
            def __init__(self):
                self.api_key = 'YOU FELL FOR IT FOOL'

        tv = 'after-init'
        test = Test()
        assert test.api_key != tv, 'should have failed due to instance setting its own api_key'

    def test_after_init_adt_unset(self):
        dec = self.auth.tangential('api_key', 'test-after-init', atInit=True)
        @dec(atInit=False)
        class Test:
            pass

        tv = 'after-init'
        assert Test.api_key == tv
        test = Test()
        assert test.api_key == tv, [d for d in test.__dict__ if not d.startswith('__')]

    def test_tangential_init(self):
        @self.auth.tangential_init('api_key', 'test-tang-init')
        class Test:
            pass

        assert not hasattr(Test, 'api_key')
        tv = 'tiapi'
        test = Test()
        assert test.api_key == tv, [d for d in test.__dict__ if not d.startswith('__')]

    def test_tangential_init_after(self):
        @self.auth.tangential_init('api_key', 'test-after-init', after=True)
        class Test:
            def __init__(self):
                self.api_key = 'YOU FELL FOR IT FOOL'

        assert not hasattr(Test, 'api_key')
        tv = 'after-init'
        test = Test()
        assert test.api_key == tv, [d for d in test.__dict__ if not d.startswith('__')]

    def test_multi(self):
        @self.auth.tangential('v1', 'test-value-1', atInit=True)
        @self.auth.tangential('v2', 'test-value-2', asProperty=True)
        @self.auth.tangential_init('api_key', 'test-tang-init')
        class Test:
            pass

        test = Test()

        assert not hasattr(Test, 'v1')
        assert hasattr(Test, 'v2')
        assert Test.v2, [d for d in dir(Test) if not d.startswith('__')]
        assert not hasattr(Test, 'api_key')

        assert test.v1 == 'a auth-config value', [d for d in test.__dict__ if not d.startswith('__')]
        assert test.v2 == 'a user value', [d for d in test.__dict__ if not d.startswith('__')]
        assert test.api_key == 'tiapi'



import os
import shutil
import unittest
import pytest
import orthauth as oa
from .common import test_folder


class TestAltConfig(unittest.TestCase):
    single = test_folder / 'auth-config-1.yaml'
    top = test_folder / 'auth-config-5.yaml'
    top_rename = test_folder / 'auth-config-6.yaml'
    def test_single(self):
        auth = oa.configure(self.single)
        tv = 'alt config single value'
        test = auth.get('test-alt-config-single')
        assert test == tv
    def test_single_rename(self):
        auth = oa.configure(self.single)
        tv = 'alt config single rename value'
        test = auth.get('test-alt-config-single-rename')
        assert test == tv
    def test_top(self):
        auth = oa.configure(self.top)
        tv = 'alt config top value'
        test = auth.get('test-top-level-alt-config')
        assert test == tv
    def test_top_rename(self):
        auth = oa.configure(self.top_rename)
        tv = 'alt config top rename value'
        test = auth.get('name-that-would-collide-or-something')
        assert test == tv


class TestInclude(unittest.TestCase):
    p1 = test_folder / 'auth-config-1.yaml'
    p2 = test_folder / 'auth-config-2.json'
    p3 = test_folder / 'auth-config-3.py'
    p4 = test_folder / 'auth-config-4.py'
    p5 = test_folder / 'one' / 'auth-config-10.yaml'
    p6 = test_folder / 'two/three' / 'auth-config-11.yaml'

    def tearDown(self):
        for p in (self.p2, self.p3, self.p4):
            a = oa.configure(p)
            ucp = a.user_config._path
            if ucp.parent != p.parent:
                shutil.rmtree(p.parent / ucp.relative_to(p.parent).parts[0])
            else:
                ucp.unlink()

    def test_include(self):
        oa.configure(self.p1, include=(self.p2,))

    def test_collision(self):
        try:
            oa.configure(self.p1, include=(self.p3,))
            raise AssertionError('should have failed due to collision')
        except oa.exceptions.VariableCollisionError:
            pass

    def test_collision_between_included(self):
        try:
            oa.configure(self.p1, include=(self.p2, self.p4))
            raise AssertionError('should have failed due to collision')
        except oa.exceptions.VariableCollisionError:
            pass

    def test_get_included(self):
        auth = oa.configure(self.p1, include=(self.p2,))
        tv = 'ok'
        test = auth.get('auth-config-2-test-value')
        assert tv == test

    def test_get_included_user(self):
        tv = 'super duper'
        auth = oa.configure(self.p1, include=(self.p2,))
        s = auth._include[0]
        uc = s.user_config
        blob = uc.load()
        blob['auth-variables']['auth-config-2-test-value'] = tv
        uc.dump(blob)
        test = auth.get('auth-config-2-test-value')
        assert tv == test

    def test_included_relative_path(self):
        p5c = oa.configure(self.p5)
        p6c = oa.configure(self.p6, include=(self.p5,))

        test_av = 'test-include-relative-path'

        value_p5 = p5c.get_path(test_av)
        value_p6 = p6c.get_path(test_av)
        assert value_p5 == value_p6

    def test_included_relative_path_default(self):
        p5c = oa.configure(self.p5)
        p6c = oa.configure(self.p6, include=(self.p5,))
        test_av = 'test-include-relative-path-default'

        value_p5 = p5c.get_path(test_av)
        value_p6 = p6c.get_path(test_av)
        assert value_p5 == value_p6

    def test_included_relative_path_envars(self):
        p5c = oa.configure(self.p5)
        p6c = oa.configure(self.p6, include=(self.p5,))
        test_av = 'test-include-relative-path-envars'

        value_p5 = p5c.get_path(test_av)
        value_p6 = p6c.get_path(test_av)
        assert value_p5 == value_p6



import os
import stat
import pathlib
from . import exceptions as exc
from .utils import QuietDict, log, sxpr_to_python

try:
    import yaml  # FIXME DANGERZONE :/
except ImportError as e:
    pass


class Authinfo:
    def __init__(self, path):
        if isinstance(path, str):
            path = pathlib.Path(path)

        self._path = path

    def __call__(self, *names):
        raise NotImplementedError('TODO')


class Mypass:
    def __init__(self, path):
        if isinstance(path, str):
            path = pathlib.Path(path)

        self._path = path

    def __call__(self, *names):
        raise NotImplementedError('TODO')


class SshConfig:
    def __init__(self, path):
        if isinstance(path, str):
            path = pathlib.Path(path)

        self._path = path

    def __call__(self, *names):
        raise NotImplementedError('TODO')


class Secrets:
    def __init__(self, path):
        if isinstance(path, str):
            path = pathlib.Path(path)

        self._path = path
        if self.exists:
            if os.name == 'nt':
                log.warning(
                    'Make sure other users cannot read your secrets file '
                    'because right at the moment I don\'t know how to check.')
            else:
                fstat = os.stat(self._path)
                mode = oct(stat.S_IMODE(fstat.st_mode))
                if mode != '0o600' and mode != '0o700':
                    raise exc.BadFilePermissionsError(
                        f'Your secrets file {self._path} '
                        f'can be read by the whole world! {mode}')

    @property
    def filename(self):
        return self._path.as_posix()

    @property
    def exists(self):
        """ Fail early and often when missing a file that is supposed to exist """
        e = self._path.exists()
        if not e:
            raise FileNotFoundError(self._path)

        return e

    @property
    def name_id_map(self):
        # sometimes the easiest solution is just to read from disk every single time
        if self.exists:
            if self._path.suffix in ('.yaml', '.yml'):
                with open(self.filename, 'rt') as f:
                    blob = yaml.safe_load(f)
            elif self._path.suffix == '.sxpr':
                with open(self.filename, 'rt') as f:
                    blob = sxpr_to_python(f.read())
            else:
                breakpoint()
                raise ValueError('FIXME detect this statically before a runtime error')

            if blob is None:
                msg = (f'The config at {self.filename!r} '
                       'should not be empty!')
                raise exc.EmptyConfigError(msg)

            return QuietDict(blob)  # XXX NOTE quiet is not recursive

    def __call__(self, *names):
        if self.exists:
            nidm = self.name_id_map
            # NOTE under these circumstances this pattern is ok because anyone
            # or anything who can call this function can access the secrets file.
            # Normally this would be an EXTREMELY DANGEROUS PATTERN. Because short
            # secrets could be exposted by brute force, but in this case it is ok
            # because it is more important to alert the user that they have just
            # tried to use a secret as a name and that it might be in their code.
            def all_values(d):
                for v in d.values():
                    if isinstance(v, dict):
                        yield from all_values(v)
                    else:
                        yield v

            av = set(all_values(nidm))
            current = nidm
            nidm = None
            del nidm
            for name in names:
                if name in av:
                    ANGRY = '*' * len(name)
                    av = None
                    name = None
                    names = None
                    current = None
                    del av
                    del name
                    del names
                    del current
                    raise exc.SecretAsKeyError(f'WHY ARE YOU TRYING TO USE A SECRET {ANGRY} AS A NAME!?')
                elif not isinstance(current, dict):
                    av = None
                    name = None
                    names = None
                    current = None
                    del av
                    del name
                    del names
                    del current
                    # XXX reminder that this can be used for brute
                    # force to see if a terminal path value exists,
                    # one wouldn't actually do this, but be aware
                    msg = 'This secret path does not exist.'
                    raise exc.SecretPathError(msg)
                else:
                    try:
                        current = current[name]
                    except KeyError as e:
                        av = None
                        name = None
                        names = None
                        current = None
                        del av
                        del name
                        del names
                        del current
                        msg = 'This secret path does not exist.'
                        raise exc.SecretPathError(msg) from e

            if isinstance(current, dict):
                keys = sorted(current.keys())
                ANGRIER = ['*' * len(k) for k in keys if k in av]
                if ANGRIER:
                    if len(ANGRIER) == 1:
                        msg = f'WHY ARE YOU TRYING TO USE A SECRET {ANGRIER[0]} AS A NAME!?'
                    else:
                        msg = ('WHY ARE YOU TRYING TO USE MULTIPLE (!!) '
                               f'SECRETS {" ".join(ANGRIER)} AS A NAME!?')

                    av = None
                    name = None
                    names = None
                    current = None
                    keys = None
                    del av
                    del name
                    del names
                    del current
                    del keys
                    raise exc.SecretAsKeyError(msg)

                else:
                    msg = f'This secret path is incomplete. Keys are {keys}'
                    av = None
                    name = None
                    names = None
                    current = None
                    del av
                    del name
                    del names
                    del current
                    raise exc.SecretPathError(msg)

            if current is None:
                # empty secrets are an error
                msg = f'Value of secret at {names} is None.'
                av = None
                name = None
                names = None
                current = None
                del av
                del name
                del names
                del current
                raise exc.SecretEmptyError(msg)

            return current


class Runtime(Secrets):

    exists = True
    _path = None

    def __init__(self, blob):
        self._sblob = QuietDict(blob)  # XXX NOTE quiet is not recursive

    @property
    def name_id_map(self):
        return self._sblob



import re
from setuptools import setup


def find_version(filename):
    _version_re = re.compile(r"__version__ = ['\"](.*)['\"]")
    last = None  # match python semantics
    for line in open(filename):
        version_match = _version_re.match(line)
        if version_match:
            last = version_match.group(1)

    return last


__version__ = find_version('orthauth/__init__.py')

with open('README.md', 'rt') as f:
    long_description = f.read()

yaml_requires = ['pyyaml']
sxpr_requires = ['sxpyr>=0.0.2']
tests_require = ['pytest', 'pytest-runner'] + yaml_requires + sxpr_requires
setup(name='orthauth',
      version=__version__,
      description='A library to separate configuration and authentication from program logic',
      long_description=long_description,
      long_description_content_type='text/markdown',
      url='https://github.com/tgbugs/orthauth',
      author='Tom Gillespie',
      author_email='tgbugs@gmail.com',
      license='MIT',
      classifiers=[
          'Development Status :: 4 - Beta',
          'License :: OSI Approved :: MIT License',
          'Programming Language :: Python :: 3.6',
          'Programming Language :: Python :: 3.7',
          'Programming Language :: Python :: 3.8',
          'Programming Language :: Python :: 3.9',
          'Programming Language :: Python :: 3.10',
          'Programming Language :: Python :: 3.11',
          'Programming Language :: Python :: Implementation :: CPython',
          'Programming Language :: Python :: Implementation :: PyPy',
          'Operating System :: POSIX :: Linux',
          'Operating System :: MacOS :: MacOS X',
          'Operating System :: Microsoft :: Windows',
      ],
      keywords=('python orthogonal authentication '
                'config configuration management'),
      packages=[
          'orthauth',
      ],
      python_requires='>=3.6',
      tests_require=tests_require,
      install_requires=[],
      extras_require={'test': tests_require,
                      'yaml': yaml_requires,
                      'sxpr': sxpr_requires,
                     },
      scripts=[],
      entry_points={'console_scripts': [ ],},
     )



import os
import sys
import pathlib
import unittest
import yaml
import orthauth as oa
from .common import test_folder, s1, s2


class TestDeprecations(unittest.TestCase):
    def test_d_dynamic_config(self):
        auth = oa.configure_here('auth-config.py', __name__)
        dc = auth.dynamic_config


class TestAuthConfig(unittest.TestCase):
    def test_auth_config_in_binary_blob(self):
        auth = oa.configure_here('auth-config.py', __name__)
        blob = auth.load()
        assert blob == {'config-search-paths': ['../test/configs/user-1.yaml'],
                        'auth-variables': {'hrm': 'derp'}}

    def test_bad_alt_config(self):
        try:
            auth = oa.configure(test_folder / 'auth-config-bad-alt-config.yaml')
            auth.get('test')
            raise AssertionError('should have failed')
        except oa.exceptions.BadAuthConfigFormatError:
            pass


class TestConfigure(unittest.TestCase):
    def setUp(self):
        with open(test_folder / 'auth-config-1.yaml', 'rt') as f:
            self.tv = yaml.safe_load(f)

    def test_configure(self):
        auth = oa.configure(test_folder / 'auth-config-1.yaml')
        test = auth.load()
        assert test == self.tv

    def test_configure_here(self):
        auth = oa.configure_here('auth-config-0.yaml', __name__)
        test = auth.load()
        tv = {'config-search-paths': ['configs/user-1.yaml'],
              'auth-variables': {'default-example': None}}
        assert test == tv

    def test_configure_relative(self):
        auth = oa.configure_relative('auth-config-0.yaml')
        test = auth.load()
        tv = {'config-search-paths': ['configs/user-1.yaml'],
              'auth-variables': {'default-example': None}}
        assert test == tv


class TestSimple(unittest.TestCase):
    def setUp(self):
        sc = test_folder / 'auth-config-1.yaml'
        self.auth = oa.AuthConfig(sc)

    def test_null_path(self):
        assert self.auth.get('test-null-path') is None
        assert self.auth.get_path('test-null-path') is None

    def test_config_vars(self):
        g = self.auth.get('test-config-vars')
        d = self.auth.get_default('test-config-vars')
        assert g != d
        gtv = self.auth._pathit(g)
        dtv = self.auth._pathit(d)
        assert gtv != dtv

        test = self.auth.get_path('test-config-vars')
        assert test == pathlib.Path(sys.prefix, 'share/orthauth/.does-not-exist')
        assert test == gtv
        assert test != dtv

    def test_get_default(self):
        g = self.auth.get('test-config-vars')
        d = self.auth.get_default('test-config-vars')
        assert g != d
        gtv = self.auth._pathit(g)
        dtv = self.auth._pathit(d)
        assert gtv != dtv

        test = self.auth.get_path_default('test-config-vars')
        assert test == pathlib.Path.cwd() / 'share/orthauth/.does-not-exist'
        assert test != gtv
        assert test == dtv

    def test_user(self):
        print(self.auth)
        assert self.auth.user_config._path == self.auth.user_config_path, 'big oops'

    def test_tangential(self):
        @self.auth.tangential('api_key', 'full-complexity-example')
        class Test:
            pass

        assert Test.api_key, [d for d in dir(Test) if not d.startswith('__')]

    def test_tangential_env_none(self):
        @self.auth.tangential('should_be_none', 'test-env-none')
        class Test:
            pass

        assert Test.should_be_none is None

    def test_tangential_init_env_none(self):
        @self.auth.tangential_init('should_be_none', 'test-env-none')
        class Test:
            pass

        t = Test()
        assert t.should_be_none is None

    def test_path_list_variant(self):
        assert self.auth.get('paths-as-list-example') == 'lol'

    def test_path_strings(self):
        assert self.auth.get('paths-example') == 'yay!'

    def test_implicit_env(self):
        assert self.auth.get('env-example') == os.environ.get('USER', None)

    def test_env_list(self):
        tv = 'so'
        os.environ['QUITE'] = tv
        try:
            assert self.auth.get('env-list-example') == tv
        finally:
            os.environ.pop('QUITE')

    def test_env_none(self):
        assert self.auth.get('test-env-none') is None

    def test_expanduser(self):
        tv = pathlib.Path('~/').expanduser()
        assert self.auth.get_path('test-expanduser') == tv

    def test_default(self):
        assert self.auth.get('default-example') == 42

    def test_multi_path_1(self):
        assert self.auth.get_path('test-multi-path-1') == self.auth._path

    def test_multi_path_2(self):
        assert self.auth.get_path('test-multi-path-2') == self.auth.user_config._path

    def test_multi_path_3(self):
        assert self.auth.get_path('test-multi-path-3') == self.auth.user_config._path

    def test_get_list(self):
        tv = self.auth.get_list('test-get-list')
        assert tv == ['this', 'is', 'a', 'list', 'from', 'user', 'config']

    def test_get_list_default(self):
        tv = self.auth.get_list('test-get-list-default')
        assert tv == ['this', 'is', 'a', 'list', 'default']

    def test_get_list_empty(self):
        tv = self.auth.get_list('test-get-list-empty')
        assert isinstance(tv, list) and not tv

    def test_get_list_not_in_user_config(self):
        tv = self.auth.get_list('test-get-list-not-in-user-config')
        assert isinstance(tv, list) and not tv

    def test_get_list_default_and_user(self):
        tv = self.auth.get_list('test-get-list-default-and-user')
        assert tv == ['user']

    def test_key_only_in_user_config(self):
        # FIXME the issue only surfaces if the value is None which is bad
        tv_nn = self.auth.get('test-key-only-in-user-config-not-null')
        assert tv_nn == 'hello'
        # we need to warn on this ?? I'm pretty sure that I have made
        # use of the ability to do this at some point in time to issue
        # a patch fix or something like that, might be worth keeping for
        # that reason?
        try:
            tv = self.auth.get('test-key-only-in-user-config')
            # FIXME not sure if correct behavior, pretty sure
            # we don't want this to fall through here but
            # instead want it to be an error because it is
            # something can only happen during development
            # since it requires the code to know about an
            # auth variable that the auth config does not contain
            # FIXME we may have to do this in the linting pass
            assert tv == None
            assert False, 'should not get here? see comments'
        except KeyError as e:
            pass


class TestMakeUserConfig(unittest.TestCase):
    def setUp(self):
        sc = test_folder / 'auth-config-1.yaml'
        self.auth = oa.AuthConfig(sc)

    def test_exists(self):
        try:
            self.auth.write_user_config()
            raise AssertionError('should not have written')
        except oa.exceptions.ConfigExistsError:
            pass

    def test_make_user(self):
        output = self.auth._make_user()
        assert output['auth-variables'].keys() == self.auth.get_blob('auth-variables').keys(), 'hrm'

    def _roundtrip(self, format):
        ser = self.auth._serialize_user_config(format=format)
        reloaded = self.auth._load_string(ser, format)
        test = self.auth._make_user()
        assert reloaded == test, 'roundtrip failed for {format}'

    def test_serialize_json(self):
        self._roundtrip('json')

    def test_serialize_py(self):
        self._roundtrip('py')

    def test_serialize_yaml(self):
        self._roundtrip('yaml')

    def test_serialize_sxpr(self):
        self._roundtrip('sxpr')


class TestWithStores(unittest.TestCase):
    def setUp(self):
        sc = test_folder / 'auth-config-8.yaml'
        self.auth = oa.AuthConfig(sc)

    def test_path_relative_store(self):
        tv = s2.parent / 'some-other-relative-path.ext'
        test = self.auth.get_path('rel-secrets-path')
        assert test == tv


class TestPathInConfig(unittest.TestCase):
    def setUp(self):
        sc = test_folder / 'auth-config-9.yaml'
        self.auth = oa.AuthConfig(sc)

    def test_path_in_user_config(self):
        tv = test_folder / 'somewhere-else' / 'some-other-file.ext'
        test = self.auth.get_path('path-in-user-config')
        assert test == tv

    def test_paths_in_user_config(self):
        tv = test_folder / 'somewhere-else' / 'some-other-file.ext'
        test = self.auth.get_path('paths-in-user-config')
        assert test == tv


class TestConfigPathNoKeyFilePath(unittest.TestCase):
    def setUp(self):
        sc = test_folder / 'auth-get-path.yaml'
        self.auth = oa.AuthConfig(sc)

    def test_get_file_path_from_config_path(self):
        # the case where a default value is a valid path in a secrets
        # files could be added to linting, but cannot be default
        # behavior due to ambiguity
        try:
            test = self.auth.get_path('some-path-var')
            assert False, 'should have failed'
        except oa.exceptions.SomethingWrongWithVariableInConfig as e:
            pass

    def test_config_path_ok(self):
        tv = test_folder / 'somewhere-else' / 'some-other-file.ext'
        test = self.auth.get_path('spv-1')
        assert test == tv

    def test_config_path_missing(self):
        try:
            test = self.auth.get_path('spv-2')
            assert False, 'should have failed'
        except oa.exceptions.SecretError as e:
            pass

    def test_config_path_beyond(self):
        try:
            test = self.auth.get_path('spv-3')
            assert False, 'should have failed'
        except oa.exceptions.SecretError as e:
            pass


class TestBadPaths(unittest.TestCase):
    def setUp(self):
        sc = test_folder / 'auth-config-bad-paths.yaml'
        self.auth = oa.AuthConfig(sc)

    def test_path_relative_store(self):
        try:
            test = self.auth.get_path('test')
            assert False, 'should have failed'
        except oa.exceptions.SomethingWrongWithVariableInConfig as e:
            print(e)


class TestEmptySecrets(unittest.TestCase):
    def setUp(self):
        sc = test_folder / 'auth-config-empty-secrets.yaml'
        self.auth = oa.AuthConfig(sc)

    def test_empty_path(self):
        try:
            test = self.auth.get_path('missing')
            assert False, 'should have failed'
        except oa.exceptions.EmptyConfigError as e:
            print(e)

    def test_empty_value(self):
        try:
            test = self.auth.get('missing')
            assert False, 'should have failed'
        except oa.exceptions.EmptyConfigError as e:
            print(e)
